from machine import I2C
import time
import math

GRAVITY_EARTH = 9.8
gyroscopeOffset = [0, 0, 0]
magnetometerData = [0, 0, 0, 0, 0, 0, 0, 0]
magnetometerTempX = [0, 0, 0, 0, 0, 0, 0, 0, 0]
magnetometerTempY = [0, 0, 0, 0, 0, 0, 0, 0, 0]
magnetometerTempZ = [0, 0, 0, 0, 0, 0, 0, 0, 0]

Ki = 1.0
Kp = 4.50

q0 = 1.0
q1 = 0.0
q2 = 0.0
q3 = 0.0

# define ICM-20948
I2C_ADD_ICM20948 = 0x68
I2C_ADD_ICM20948_AK09916 = 0x0C
I2C_ADD_ICM20948_AK09916_READ = 0x80
I2C_ADD_ICM20948_AK09916_WRITE = 0x00

# define ICM-20948 Register
# User bank 0 register
REG_ADD_WIA = 0x00
REG_VAL_WIA = 0xEA
REG_ADD_USER_CTRL = 0x03
REG_VAL_BIT_DMP_EN = 0x80
REG_VAL_BIT_FIFO_EN = 0x40
REG_VAL_BIT_I2C_MST_EN = 0x20
REG_VAL_BIT_I2C_IF_DIS = 0x10
REG_VAL_BIT_DMP_RST = 0x08
REG_VAL_BIT_DIAMOND_DMP_RST = 0x04
REG_ADD_PWR_MIGMT_1 = 0x06
REG_VAL_ALL_RGE_RESET = 0x80
REG_VAL_RUN_MODE = 0x01  # Non low-power mode
REG_ADD_LP_CONFIG = 0x05
REG_ADD_PWR_MGMT_1 = 0x06
REG_ADD_PWR_MGMT_2 = 0x07
REG_ADD_ACCEL_XOUT_H = 0x2D
REG_ADD_ACCEL_XOUT_L = 0x2E
REG_ADD_ACCEL_YOUT_H = 0x2F
REG_ADD_ACCEL_YOUT_L = 0x30
REG_ADD_ACCEL_ZOUT_H = 0x31
REG_ADD_ACCEL_ZOUT_L = 0x32
REG_ADD_GYRO_XOUT_H = 0x33
REG_ADD_GYRO_XOUT_L = 0x34
REG_ADD_GYRO_YOUT_H = 0x35
REG_ADD_GYRO_YOUT_L = 0x36
REG_ADD_GYRO_ZOUT_H = 0x37
REG_ADD_GYRO_ZOUT_L = 0x38
REG_ADD_EXT_SENS_DATA_00 = 0x3B
REG_ADD_REG_BANK_SEL = 0x7F
REG_VAL_REG_BANK_0 = 0x00
REG_VAL_REG_BANK_1 = 0x10
REG_VAL_REG_BANK_2 = 0x20
REG_VAL_REG_BANK_3 = 0x30

# User bank 1 register
# User bank 2 register
REG_ADD_GYRO_SMPLRT_DIV = 0x00
REG_ADD_GYRO_CONFIG_1 = 0x01
REG_VAL_BIT_GYRO_DLPCFG_2 = 0x10  # bit[5:3]
REG_VAL_BIT_GYRO_DLPCFG_4 = 0x20  # bit[5:3]
REG_VAL_BIT_GYRO_DLPCFG_6 = 0x30  # bit[5:3]
REG_VAL_BIT_GYRO_FS_250DPS = 0x00  # bit[2:1]
REG_VAL_BIT_GYRO_FS_500DPS = 0x02  # bit[2:1]
REG_VAL_BIT_GYRO_FS_1000DPS = 0x04  # bit[2:1]
REG_VAL_BIT_GYRO_FS_2000DPS = 0x06  # bit[2:1]
REG_VAL_BIT_GYRO_DLPF = 0x01  # bit[0]
REG_ADD_ACCEL_SMPLRT_DIV_2 = 0x11
REG_ADD_ACCEL_CONFIG = 0x14
REG_VAL_BIT_ACCEL_DLPCFG_2 = 0x10  # bit[5:3]
REG_VAL_BIT_ACCEL_DLPCFG_4 = 0x20  # bit[5:3]
REG_VAL_BIT_ACCEL_DLPCFG_6 = 0x30  # bit[5:3]
REG_VAL_BIT_ACCEL_FS_2g = 0x00  # bit[2:1]
REG_VAL_BIT_ACCEL_FS_4g = 0x02  # bit[2:1]
REG_VAL_BIT_ACCEL_FS_8g = 0x04  # bit[2:1]
REG_VAL_BIT_ACCEL_FS_16g = 0x06  # bit[2:1]
REG_VAL_BIT_ACCEL_DLPF = 0x01  # bit[0]

# User bank 3 register
REG_ADD_I2C_SLV0_ADDR = 0x03
REG_ADD_I2C_SLV0_REG = 0x04
REG_ADD_I2C_SLV0_CTRL = 0x05
REG_VAL_BIT_SLV0_EN = 0x80
REG_VAL_BIT_MASK_LEN = 0x07
REG_ADD_I2C_SLV0_DO = 0x06
REG_ADD_I2C_SLV1_ADDR = 0x07
REG_ADD_I2C_SLV1_REG = 0x08
REG_ADD_I2C_SLV1_CTRL = 0x09
REG_ADD_I2C_SLV1_DO = 0x0A

# define ICM-20948 MAG Register
REG_ADD_MAG_WIA1 = 0x00
REG_VAL_MAG_WIA1 = 0x48
REG_ADD_MAG_WIA2 = 0x01
REG_VAL_MAG_WIA2 = 0x09
REG_ADD_MAG_ST2 = 0x10
REG_ADD_MAG_DATA = 0x11
REG_ADD_MAG_CNTL2 = 0x31
REG_VAL_MAG_MODE_PD = 0x00
REG_VAL_MAG_MODE_SM = 0x01
REG_VAL_MAG_MODE_10HZ = 0x02
REG_VAL_MAG_MODE_20HZ = 0x04
REG_VAL_MAG_MODE_50HZ = 0x05
REG_VAL_MAG_MODE_100HZ = 0x08
REG_VAL_MAG_MODE_ST = 0x10


MAG_DATA_LEN = 6


class ICM20948(object):
    def __init__(self, address=I2C_ADD_ICM20948):
        self._address = address
        self._bus = I2C(1)

        # Initialization of the device multiple times
        # after power on will result in a return error
        bRet = self.imu_check()
        time.sleep(0.5)
        # User bank 0 register
        self._write_byte(REG_ADD_REG_BANK_SEL, REG_VAL_REG_BANK_0)
        self._write_byte(REG_ADD_PWR_MIGMT_1, REG_VAL_ALL_RGE_RESET)
        time.sleep(0.1)
        self._write_byte(REG_ADD_PWR_MIGMT_1, REG_VAL_RUN_MODE)
        # User bank 2 register
        self._write_byte(REG_ADD_REG_BANK_SEL, REG_VAL_REG_BANK_2)
        self._write_byte(REG_ADD_GYRO_SMPLRT_DIV, 0x07)
        self._write_byte(REG_ADD_GYRO_CONFIG_1, REG_VAL_BIT_GYRO_DLPCFG_6 |
                         REG_VAL_BIT_GYRO_FS_1000DPS | REG_VAL_BIT_GYRO_DLPF)
        self._write_byte(REG_ADD_ACCEL_SMPLRT_DIV_2,  0x07)

        self._write_byte(REG_ADD_ACCEL_CONFIG, REG_VAL_BIT_ACCEL_DLPCFG_6 |
                         REG_VAL_BIT_ACCEL_FS_2g | REG_VAL_BIT_ACCEL_DLPF)

        # User bank 0 register
        self._write_byte(REG_ADD_REG_BANK_SEL, REG_VAL_REG_BANK_0)
        time.sleep(0.1)
        self.read_gyroscope_offset()
        self.magnetometer_check()
        self._write_secondary(
            I2C_ADD_ICM20948_AK09916 | I2C_ADD_ICM20948_AK09916_WRITE, REG_ADD_MAG_CNTL2, REG_VAL_MAG_MODE_20HZ)

    def read_accelerometer(self):
        accelerometer = [0, 0, 0]
        self._write_byte(REG_ADD_REG_BANK_SEL, REG_VAL_REG_BANK_0)
        data = self._read_bytes(REG_ADD_ACCEL_XOUT_H, 6)
        self._write_byte(REG_ADD_REG_BANK_SEL, REG_VAL_REG_BANK_2)
        accelerometer[0] = (data[0] << 8) | data[1]
        accelerometer[1] = (data[2] << 8) | data[3]
        accelerometer[2] = (data[4] << 8) | data[5]
        if accelerometer[0] >= 32767:
            accelerometer[0] = accelerometer[0]-65535
        elif accelerometer[0] <= -32767:
            accelerometer[0] = accelerometer[0]+65535
        if accelerometer[1] >= 32767:
            accelerometer[1] = accelerometer[1]-65535
        elif accelerometer[1] <= -32767:
            accelerometer[1] = accelerometer[1]+65535
        if accelerometer[2] >= 32767:
            accelerometer[2] = accelerometer[2]-65535
        elif accelerometer[2] <= -32767:
            accelerometer[2] = accelerometer[2]+65535
        return accelerometer

    def read_accelerometer_g(self):
        accelerometer_g = [0,0,0]
        accelerometer = self.read_accelerometer()
        accelerometer_g[0] = accelerometer[0] * 2.0 / 32768.0
        accelerometer_g[1] = accelerometer[1] * 2.0 / 32768.0
        accelerometer_g[2] = accelerometer[2] * 2.0 / 32768.0
        accelerometer_g[0] = round(accelerometer_g[0], 2)
        accelerometer_g[1] = round(accelerometer_g[1], 2)
        accelerometer_g[2] = round(accelerometer_g[2], 2)
        return accelerometer_g
    
    def read_accelerometer_a(self):
        accelerometer_a = [0,0,0]
        accelerometer_g = self.read_accelerometer_g()
        accelerometer_a[0] = accelerometer_g[0] * GRAVITY_EARTH
        accelerometer_a[1] = accelerometer_g[1] * GRAVITY_EARTH
        accelerometer_a[2] = accelerometer_g[2] * GRAVITY_EARTH
        return accelerometer_a
        
    def read_gyroscope(self):
        gyroscope = [0, 0, 0]
        self._write_byte(REG_ADD_REG_BANK_SEL, REG_VAL_REG_BANK_0)
        data = self._read_bytes(REG_ADD_GYRO_XOUT_H, 6)
        self._write_byte(REG_ADD_REG_BANK_SEL, REG_VAL_REG_BANK_2)
        gyroscope[0] = ((data[0] << 8) | data[1]) - gyroscopeOffset[0]
        gyroscope[1] = ((data[2] << 8) | data[3]) - gyroscopeOffset[1]
        gyroscope[2] = ((data[4] << 8) | data[5]) - gyroscopeOffset[2]
        if gyroscope[0] >= 32767:
            gyroscope[0] = gyroscope[0]-65535
        elif gyroscope[0] <= -32767:
            gyroscope[0] = gyroscope[0]+65535
        if gyroscope[1] >= 32767:
            gyroscope[1] = gyroscope[1]-65535
        elif gyroscope[1] <= -32767:
            gyroscope[1] = gyroscope[1]+65535
        if gyroscope[2] >= 32767:
            gyroscope[2] = gyroscope[2]-65535
        elif gyroscope[2] <= -32767:
            gyroscope[2] = gyroscope[2]+65535
        return gyroscope

    def read_gyroscope_deg(self):
        gyroscope_deg = [0,0,0]
        gyroscope = self.read_gyroscope()
        gyroscope_deg[0] = gyroscope[0] * 250.0 / 32768.0
        gyroscope_deg[1] = gyroscope[1] * 250.0 / 32768.0
        gyroscope_deg[2] = gyroscope[2] * 250.0 / 32768.0
        gyroscope_deg[0] = round(gyroscope_deg[0], 2)
        gyroscope_deg[1] = round(gyroscope_deg[1], 2)
        gyroscope_deg[2] = round(gyroscope_deg[2], 2)
        return gyroscope_deg
    
    def read_gyroscope_rad(self):
        gyroscope_rad = [0,0,0]
        gyroscope_deg = self.read_gyroscope_deg()
        gyroscope_rad[0] = math.radians(gyroscope_deg[0])
        gyroscope_rad[1] = math.radians(gyroscope_deg[1])
        gyroscope_rad[2] = math.radians(gyroscope_deg[2])
        gyroscope_rad[0] = round(gyroscope_rad[0], 2)
        gyroscope_rad[1] = round(gyroscope_rad[1], 2)
        gyroscope_rad[2] = round(gyroscope_rad[2], 2)
        return gyroscope_rad

    def read_magnetometer(self):
        magnetometer = [0, 0, 0]
        counter = 20
        while(counter > 0):
            time.sleep(0.01)
            self._read_secondary(
                I2C_ADD_ICM20948_AK09916 | I2C_ADD_ICM20948_AK09916_READ, REG_ADD_MAG_ST2, 1)
            if ((magnetometerData[0] & 0x01) != 0):
                break
            counter -= 1
        if counter != 0:
            for i in range(0, 8):
                self._read_secondary(
                    I2C_ADD_ICM20948_AK09916 | I2C_ADD_ICM20948_AK09916_READ, REG_ADD_MAG_DATA, MAG_DATA_LEN)
                magnetometerTempX[i] = (
                    magnetometerData[1] << 8) | magnetometerData[0]
                magnetometerTempY[i] = (
                    magnetometerData[3] << 8) | magnetometerData[2]
                magnetometerTempZ[i] = (
                    magnetometerData[5] << 8) | magnetometerData[4]
            magnetometer[0] = (magnetometerTempX[0]+magnetometerTempX[1]+magnetometerTempX[2]+magnetometerTempX[3] +
                               magnetometerTempX[4]+magnetometerTempX[5]+magnetometerTempX[6]+magnetometerTempX[7])/8
            magnetometer[1] = -(magnetometerTempY[0]+magnetometerTempY[1]+magnetometerTempY[2]+magnetometerTempY[3] +
                                magnetometerTempY[4]+magnetometerTempY[5]+magnetometerTempY[6]+magnetometerTempY[7])/8
            magnetometer[2] = -(magnetometerTempZ[0]+magnetometerTempZ[1]+magnetometerTempZ[2]+magnetometerTempZ[3] +
                                magnetometerTempZ[4]+magnetometerTempZ[5]+magnetometerTempZ[6]+magnetometerTempZ[7])/8
        # Solve the problem that Python shift will not overflow
        if magnetometer[0] >= 32767:
            magnetometer[0] = magnetometer[0]-65535
        elif magnetometer[0] <= -32767:
            magnetometer[0] = magnetometer[0]+65535
        if magnetometer[1] >= 32767:
            magnetometer[1] = magnetometer[1]-65535
        elif magnetometer[1] <= -32767:
            magnetometer[1] = magnetometer[1]+65535
        if magnetometer[2] >= 32767:
            magnetometer[2] = magnetometer[2]-65535
        elif magnetometer[2] <= -32767:
            magnetometer[2] = magnetometer[2]+65535

        magnetometer[0] = int(magnetometer[0])
        magnetometer[1] = int(magnetometer[1])
        magnetometer[2] = int(magnetometer[2])
        return magnetometer

    def read_magnetometer_uT(self):
        magnetometer_uT = [0,0,0]
        magnetometer = self.read_magnetometer()
        magnetometer_uT[0] = magnetometer[0] * 4912.0 / 32752.0
        magnetometer_uT[1] = magnetometer[1] * 4912.0 / 32752.0
        magnetometer_uT[2] = magnetometer[2] * 4912.0 / 32752.0
        magnetometer_uT[0] = int(magnetometer_uT[0])
        magnetometer_uT[1] = int(magnetometer_uT[1])
        magnetometer_uT[2] = int(magnetometer_uT[2])
        return magnetometer_uT
    
    def _read_secondary(self, u8I2CAddr, u8RegAddr, u8Len):
        u8Temp = 0
        self._write_byte(REG_ADD_REG_BANK_SEL,
                         REG_VAL_REG_BANK_3)
        self._write_byte(REG_ADD_I2C_SLV0_ADDR, u8I2CAddr)
        self._write_byte(REG_ADD_I2C_SLV0_REG,  u8RegAddr)
        self._write_byte(REG_ADD_I2C_SLV0_CTRL, REG_VAL_BIT_SLV0_EN | u8Len)
        self._write_byte(REG_ADD_REG_BANK_SEL,
                         REG_VAL_REG_BANK_0)

        u8Temp = self._read_byte(REG_ADD_USER_CTRL)
        u8Temp |= REG_VAL_BIT_I2C_MST_EN
        self._write_byte(REG_ADD_USER_CTRL, u8Temp)
        time.sleep(0.01)
        u8Temp &= ~REG_VAL_BIT_I2C_MST_EN
        self._write_byte(REG_ADD_USER_CTRL, u8Temp)

        for i in range(0, u8Len):
            magnetometerData[i] = self._read_byte(REG_ADD_EXT_SENS_DATA_00+i)
        self._write_byte(REG_ADD_REG_BANK_SEL,
                         REG_VAL_REG_BANK_3)

        u8Temp = self._read_byte(REG_ADD_I2C_SLV0_CTRL)
        u8Temp &= ~((REG_VAL_BIT_I2C_MST_EN) & (REG_VAL_BIT_MASK_LEN))
        self._write_byte(REG_ADD_I2C_SLV0_CTRL,  u8Temp)

        self._write_byte(REG_ADD_REG_BANK_SEL,
                         REG_VAL_REG_BANK_0)

    def _write_secondary(self, u8I2CAddr, u8RegAddr, u8data):
        u8Temp = 0
        self._write_byte(REG_ADD_REG_BANK_SEL,
                         REG_VAL_REG_BANK_3)
        self._write_byte(REG_ADD_I2C_SLV1_ADDR, u8I2CAddr)
        self._write_byte(REG_ADD_I2C_SLV1_REG,  u8RegAddr)
        self._write_byte(REG_ADD_I2C_SLV1_DO,   u8data)
        self._write_byte(REG_ADD_I2C_SLV1_CTRL, REG_VAL_BIT_SLV0_EN | 1)

        self._write_byte(REG_ADD_REG_BANK_SEL,
                         REG_VAL_REG_BANK_0)

        u8Temp = self._read_byte(REG_ADD_USER_CTRL)
        u8Temp |= REG_VAL_BIT_I2C_MST_EN
        self._write_byte(REG_ADD_USER_CTRL, u8Temp)
        time.sleep(0.01)
        u8Temp &= ~REG_VAL_BIT_I2C_MST_EN
        self._write_byte(REG_ADD_USER_CTRL, u8Temp)

        self._write_byte(REG_ADD_REG_BANK_SEL,
                         REG_VAL_REG_BANK_3)

        u8Temp = self._read_byte(REG_ADD_I2C_SLV0_CTRL)
        u8Temp &= ~((REG_VAL_BIT_I2C_MST_EN) & (REG_VAL_BIT_MASK_LEN))
        self._write_byte(REG_ADD_I2C_SLV0_CTRL,  u8Temp)

        self._write_byte(REG_ADD_REG_BANK_SEL,
                         REG_VAL_REG_BANK_0)

    def read_gyroscope_offset(self):
        s32TempGx = 0
        s32TempGy = 0
        s32TempGz = 0
        for i in range(0, 32):
            gyroscope = self.read_gyroscope()
            s32TempGx += gyroscope[0]
            s32TempGy += gyroscope[1]
            s32TempGz += gyroscope[2]
            time.sleep(0.01)
        gyroscopeOffset[0] = s32TempGx >> 5
        gyroscopeOffset[1] = s32TempGy >> 5
        gyroscopeOffset[2] = s32TempGz >> 5

    def imu_check(self):
        bRet = False
        if REG_VAL_WIA == self._read_byte(REG_ADD_WIA):
            bRet = True
        return bRet

    def magnetometer_check(self):
        self._read_secondary(
            I2C_ADD_ICM20948_AK09916 | I2C_ADD_ICM20948_AK09916_READ, REG_ADD_MAG_WIA1, 2)
        if (magnetometerData[0] == REG_VAL_MAG_WIA1) and (magnetometerData[1] == REG_VAL_MAG_WIA2):
            bRet = True
            return bRet

    def AHRS_update(self, gx, gy, gz, ax, ay, az, mx, my, mz):
        norm = 0.0
        hx = hy = hz = bx = bz = 0.0
        vx = vy = vz = wx = wy = wz = 0.0
        exInt = eyInt = ezInt = 0.0
        ex = ey = ez = 0.0
        halfT = 0.024
        global q0
        global q1
        global q2
        global q3
        q0q0 = q0 * q0
        q0q1 = q0 * q1
        q0q2 = q0 * q2
        q0q3 = q0 * q3
        q1q1 = q1 * q1
        q1q2 = q1 * q2
        q1q3 = q1 * q3
        q2q2 = q2 * q2
        q2q3 = q2 * q3
        q3q3 = q3 * q3
        norm = float(1/math.sqrt(ax * ax + ay * ay + az * az))
        ax = ax * norm
        ay = ay * norm
        az = az * norm

        norm = float(1/math.sqrt(mx * mx + my * my + mz * mz))
        mx = mx * norm
        my = my * norm
        mz = mz * norm

        # Compute reference direction of flux
        hx = 2 * mx * (0.5 - q2q2 - q3q3) + 2 * my * \
            (q1q2 - q0q3) + 2 * mz * (q1q3 + q0q2)
        hy = 2 * mx * (q1q2 + q0q3) + 2 * my * \
            (0.5 - q1q1 - q3q3) + 2 * mz * (q2q3 - q0q1)
        hz = 2 * mx * (q1q3 - q0q2) + 2 * my * (q2q3 + q0q1) + \
            2 * mz * (0.5 - q1q1 - q2q2)
        bx = math.sqrt((hx * hx) + (hy * hy))
        bz = hz

        # Estimated direction of gravity and flux (v and w)
        vx = 2 * (q1q3 - q0q2)
        vy = 2 * (q0q1 + q2q3)
        vz = q0q0 - q1q1 - q2q2 + q3q3
        wx = 2 * bx * (0.5 - q2q2 - q3q3) + 2 * bz * (q1q3 - q0q2)
        wy = 2 * bx * (q1q2 - q0q3) + 2 * bz * (q0q1 + q2q3)
        wz = 2 * bx * (q0q2 + q1q3) + 2 * bz * (0.5 - q1q1 - q2q2)

        # Error is sum of cross product between reference direction of fields and direction measured by sensors
        ex = (ay * vz - az * vy) + (my * wz - mz * wy)
        ey = (az * vx - ax * vz) + (mz * wx - mx * wz)
        ez = (ax * vy - ay * vx) + (mx * wy - my * wx)

        if (ex != 0.0 and ey != 0.0 and ez != 0.0):
            exInt = exInt + ex * Ki * halfT
            eyInt = eyInt + ey * Ki * halfT
            ezInt = ezInt + ez * Ki * halfT

            gx = gx + Kp * ex + exInt
            gy = gy + Kp * ey + eyInt
            gz = gz + Kp * ez + ezInt

        q0 = q0 + (-q1 * gx - q2 * gy - q3 * gz) * halfT
        q1 = q1 + (q0 * gx + q2 * gz - q3 * gy) * halfT
        q2 = q2 + (q0 * gy - q1 * gz + q3 * gx) * halfT
        q3 = q3 + (q0 * gz + q1 * gy - q2 * gx) * halfT

        norm = float(1/math.sqrt(q0 * q0 + q1 * q1 + q2 * q2 + q3 * q3))
        q0 = q0 * norm
        q1 = q1 * norm
        q2 = q2 * norm
        q3 = q3 * norm
    
    def read_quaternion(self):
        quaternion = [q0, q1, q2, q3]
        return quaternion
 
    def get_pitch(self):
        pitch = math.asin(-2 * q1 * q3 + 2 * q0* q2)* 57.3
        return pitch
        
    def get_roll(self):
        roll  = math.atan2(2 * q2 * q3 + 2 * q0 * q1, -2 * q1 * q1 - 2 * q2* q2 + 1)* 57.3
        return roll

    def get_yaw(self):      
        yaw   = math.atan2(-2 * q1 * q2 - 2 * q0 * q3, 2 * q2 * q2 + 2 * q3 * q3 - 1) * 57.3
        return yaw

    def _read_byte(self, cmd):
        rec = self._bus.readfrom_mem(int(self._address), int(cmd), 1)
        return rec[0]

    def _read_byte16(self, cmd):
        LSB = self._bus.readfrom_mem(int(self._address), int(cmd), 1)
        MSB = self._bus.readfrom_mem(int(self._address), int(cmd)+1, 1)
        return (MSB[0] << 8) + LSB[0]

    def _read_bytes(self, reg, length=1):
        rec = self._bus.readfrom_mem(int(self._address), int(reg), length)
        return rec

    def _write_byte(self, cmd, val):
        self._bus.writeto_mem(int(self._address), int(cmd), bytes([int(val)]))
        time.sleep(0.0001)
